# This script merges the output trace files from the multi-threaded vehave into a single trace file.
# The script takes as an input 4  arguments: <traces_directory> <PID_of_input_file> <output_directory> <output_file_name>
# Both directories are relative paths from the location of the script.
# The script can also merge all trace files with the same PID in <traces_directory> into one file.
# To do that, pass "--all" as <PID_of_input_file> and omit <output_file_name>.


from os import listdir
from os import getcwd
from os import makedirs
from os.path import isfile, join, exists
import sys
import csv



def print_usage():
    print("Usage 3 parameters: python3 " + sys.argv[0] + " <traces_directory> --all <output_directory>")
    print("OR")
    print("Usage 4 parameters: python3 " + sys.argv[0] + " <traces_directory> <PID_of_input_file> <output_directory> <output_file_name>")
    exit()


#####################
# Preparation phase #
#####################

# Check the input arguments
argc = len(sys.argv)
if (argc != 4 and argc != 5):
    print("ERROR: 3 or 4 parameters expected, " + str(argc-1) + " found.")
    print_usage()

elif (argc == 4 and sys.argv[2] != "--all"):
    print("ERROR: unrecognized option: " + sys.argv[2])
    print_usage()

elif (argc == 5 and not str.isdigit(sys.argv[2])):
    print("ERROR: PID must be an integer number!")
    #print_usage()


input_dir = sys.argv[1]
output_dir = sys.argv[3]
#input_pid = []
#output_file_name = []

if(argc == 5):
    input_pid = sys.argv[2]
    output_file_name = sys.argv[4]

# Get current working directory
cwd = getcwd()

# Check the existence of the input and output directories
if not exists(input_dir):
    print("ERROR: can't find this directory: " + join(cwd,input_dir))
    exit()
traces_path = join(cwd,input_dir)

if not exists(output_dir):
    print("ERROR: can't find this directory: " + join(cwd,output_dir))
    exit()
output_path = join(cwd,output_dir)

# Get list of files in the input directory which starts with 'vehave.'
all_files = [f for f in listdir(traces_path) if isfile(join(traces_path, f)) and f.startswith('vehave.')]

if (argc == 5):
    file_list = []
    for f in all_files:
        if( input_pid in f):
            file_list.append(f)
    if len(file_list) == 0:
        print("ERROR: can't find trace files for PID " + str(input_pid) + " in " + traces_path)
        exit()

else:
    file_list = all_files


# Extract PID from file name and count how many files share the same PID
pid_list = []
pid_count = []
tid_list = []
for f in file_list:
    words = f.split(".")

    if '_' in words[1]:
        tmp = words[1].split("_")
        pid = tmp[0]
        tid = tmp[1]
        if pid in pid_list:
            pid_count[pid_list.index(pid)] += 1
            tid_list_of_pid = tid_list[pid_list.index(pid)]
            tid_list_of_pid.append(tid)
        else:
            pid_list.append(pid)
            pid_count.append(1)
            tid_list.append([tid])

# Sort tids for each pid
for x in tid_list:
    x.sort()


#####################
# Merge files phase #
#####################

name_prefix = "vehave."
name_suffix = ".trace"

THREAD_ID_INDEX = 9
CORE_ID_INDEX = 10

pid_list_size = len(pid_list)
for i in range(0 , pid_list_size):
    # Reading the input files
    in_file = [[] for i in range(pid_count[i])]
    in_file_len = []
    for j in range(0,pid_count[i]):
        file_name = name_prefix + str(pid_list[i]) + "_" + str(tid_list[i][j]) + name_suffix
        with open(join(traces_path, file_name),'r') as f:
            reader = csv.reader(f)
            for row in reader:
                in_file[j].append(row)
        # skip the file if empty
        if not in_file[j]:
            print("file " + str(pid_list[i]) + "_" + str(tid_list[i][j]) + " is empty!");
            print("skipping " + str(pid_list[i]));
            break;

        # find the length of the file
        in_file_len.append(len(in_file[j]))

        # Change tid to local thread number
        for l in range(1,in_file_len[-1]):
            in_file[j][l][THREAD_ID_INDEX] = tid_list[i].index(in_file[j][l][THREAD_ID_INDEX])

    # find max file length
    max_l = max(in_file_len)

    # Write the output file
    if in_file_len[0]:
        if argc == 4:
            output_file_name = name_prefix + str(pid_list[i]) + name_suffix
        with open(join(output_path, output_file_name),'w') as f:
            writer = csv.writer(f)
            writer.writerow(in_file[0][0])

            for l in range(1,max_l):
                for j in range(0,pid_count[i]):
                    if l < in_file_len[j]:
                        writer.writerow(in_file[j][l])
        if argc == 5:
            print("Merge complete for PID " + str(input_pid) + "!")
        else:
            print("Merge complete for PID " + str(pid_list[i]) + "!")
